"""Pytest fixtures and utilities for testing code that uses pytroll watchers."""

import datetime
import os
from contextlib import contextmanager, nullcontext

import pytest
import responses
import yaml


@pytest.fixture(autouse=True)
def patched_local_events(monkeypatch):
    """Patch the events produced by underlying os/polling watcher.

    Example:
        The produced context managed can be used like this:

        >>> with patched_local_events(["/tmp/file1", "/tmp/file2"]):
        ...    assert "/tmp/file1" in local_watcher.file_generator("/tmp")

    """
    @contextmanager
    def _patched_local_events(paths):
        def fake_iterator(_):
            return paths
        from pytroll_watchers.backends import local
        monkeypatch.setattr(local, "_iterate_over_queue", fake_iterator)
        yield
    return _patched_local_events


@pytest.fixture(autouse=True)
def patched_bucket_listener(monkeypatch):
    """Patch the records produced by the underlying bucket listener.

    Example:
        This context manager can be used like this:

        >>> with patched_bucket_listener(records_to_produce):
        ...     for record in bucket_notification_watcher.file_generator(endpoint, bucket):
        ...         # do something with the record

    """
    @contextmanager
    def _patched_bucket_listener(records):
        #from unittest import mock
        def fake_listen(*args, **kwargs):
            return nullcontext(enter_result=records)
        import minio
        monkeypatch.setattr(minio.Minio, "listen_bucket_notification", fake_listen)
        yield
    return _patched_bucket_listener


@contextmanager
def load_oauth_responses(*responses_to_load, response_file=None):
    """Load the oauth responses for mocking the requests to copernicus dataspace.

    Args:
        responses_to_load: The responses to load.
        response_file: The file where the responses are stored. Defaults to tests/dataspace_responses.yaml

    Example:
        To get fake response for the watcher and test the generator, one could use::

            with load_oauth_responses("token", "filtered_yesterday"):
                files = list(file_generator(filter_string, check_interval, timedelta(hours=24)))
    """
    today = datetime.datetime.now(datetime.timezone.utc)
    yesterday = today - datetime.timedelta(hours=24)

    if response_file is None:
        response_file = os.path.join("tests", "dataspace_responses.yaml")

    with responses.RequestsMock() as rsps:

        with open(response_file) as fd:
            contents = yaml.safe_load(fd.read())
        for response_to_load in responses_to_load:
            for response in contents["responses"][response_to_load]:
                resp = response["response"]
                resp["url"] = resp["url"].replace("{yesterday}", yesterday.strftime("%Y-%m-%dT%H:%M:%S.%fZ"))
                resp["url"] = resp["url"].replace("{today}", today.strftime("%Y-%m-%dT%H:%M:%S.%fZ"))
                rsps.add(**response["response"])

        yield



"""Tests for the bucket notification watcher."""

import datetime
from unittest import mock

import pytest
from posttroll.message import Message
from posttroll.testing import patched_publisher
from pytroll_watchers import minio_notification_watcher
from pytroll_watchers.publisher import SecurityError, fix_times
from pytroll_watchers.testing import patched_bucket_listener  # noqa
from upath import UPath

sdr_file_pattern = ("sdr/SV{channel_name:3s}_{platform_name}_d{start_date:%Y%m%d}_t{start_time:%H%M%S%f}_"
                    "e{end_time:%H%M%S%f}_b{orbit_number:d}_c{processing_datetime:%Y%m%d%H%M%S%f}_cspp_dev.h5")


def test_generate_paths(patched_bucket_listener):  # noqa
    """Test generating paths."""
    profile="someprofile"
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     storage_options=dict(profile=profile))
    with patched_bucket_listener(records):
        files = list(minio_notification_watcher.file_generator(**s3_config))
    assert len(files) == len(records)
    path, _ = files[0]
    assert path == UPath("s3://viirs-data/sdr/SVM13_npp_d20240408_t1006227_e1007469_b64498_c20240408102334392250_cspp_dev.h5",
                         profile=profile)

def test_generate_paths_uses_credentials_from_profile(patched_bucket_listener, monkeypatch):  # noqa
    """Test generating paths."""
    import minio
    fake_minio = mock.MagicMock()
    monkeypatch.setattr(minio, "Minio", fake_minio)

    profile="someprofile"
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     storage_options=dict(profile=profile))

    with patched_bucket_listener(records):
       _ = list(minio_notification_watcher.file_generator(**s3_config))
    assert fake_minio.mock_calls[0][2]["credentials"] is not None


def test_generate_paths_with_pattern(patched_bucket_listener):  # noqa
    """Test generating paths."""
    profile = "someprofile"
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     file_pattern=sdr_file_pattern,
                     storage_options=dict(profile=profile))
    with patched_bucket_listener(records):
        files = list(minio_notification_watcher.file_generator(**s3_config))
    assert len(files) == 10
    path, metadata = files[0]
    assert path == UPath("s3://viirs-data/sdr/SVM13_npp_d20240408_t1006227_e1007469_b64498_c20240408102334392250_cspp_dev.h5",
                         profile=profile)
    assert metadata["platform_name"] == "npp"


def test_generate_paths_with_pattern_fixes_endtime(patched_bucket_listener):  # noqa
    """Test generating paths."""
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     file_pattern=sdr_file_pattern,
                     storage_options=dict(profile="someprofile"))
    with patched_bucket_listener(records):
        files = list(minio_notification_watcher.file_generator(**s3_config))
    _, metadata = files[0]
    assert metadata["end_time"] == datetime.datetime(2024, 4, 8, 10, 7, 46, 900000)
    assert "start_date" not in metadata


def test_fix_times():
    """Test fixing times for viirs data."""
    metadata = {"channel_name": "M13",
                "platform_name": "npp",
                "start_date": datetime.datetime(2024, 4, 8, 0, 0),
                "start_time": datetime.datetime(1900, 1, 1, 23, 59, 22, 700000),
                "end_time": datetime.datetime(1900, 1, 1, 0, 0, 46, 900000),
                "orbit_number": 64498,
                "processing_datetime": datetime.datetime(2024, 4, 8, 10, 23, 34, 392250)}

    fix_times(metadata)

    assert metadata["start_time"] < metadata["end_time"]


def test_publish_paths(patched_bucket_listener, caplog):  # noqa
    """Test publishing paths."""
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     storage_options=dict(profile="someprofile"))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))

    caplog.set_level("INFO")
    with patched_publisher() as messages:
       with patched_bucket_listener(records):
              minio_notification_watcher.file_publisher(fs_config=s3_config,
                                                        publisher_config=publisher_settings,
                                                        message_config=message_settings)
    assert "uri" not in message_settings["data"]
    assert len(messages) == len(records)
    message = Message(rawstr=messages[0])
    assert message.data["uri"] == "s3://viirs-data/sdr/SVM13_npp_d20240408_t1006227_e1007469_b64498_c20240408102334392250_cspp_dev.h5"
    assert message.data["uid"] == "SVM13_npp_d20240408_t1006227_e1007469_b64498_c20240408102334392250_cspp_dev.h5"
    assert message.data["sensor"] == "viirs"
    assert message.data["filesystem"] == {"cls": "s3fs.core:S3FileSystem", "protocol": "s3", "args": [],
                                  "profile": "someprofile"}
    assert "Starting watch on 'viirs-data'" in caplog.text


def test_publish_paths_forbids_passing_secret_key(patched_bucket_listener):  # noqa
    """Test publishing paths forbids passing a secret key."""
    secret_key = "very secret"  # noqa
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     file_pattern=sdr_file_pattern,
                     storage_options=dict(access_key="my access key",
                                          secret_key=secret_key))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    with patched_publisher():
        with patched_bucket_listener(records):
            with pytest.raises(SecurityError):
                minio_notification_watcher.file_publisher(fs_config=s3_config,
                                                        publisher_config=publisher_settings,
                                                        message_config=message_settings)


def test_publish_paths_with_pattern(patched_bucket_listener):  # noqa
    """Test publishing paths."""
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     file_pattern=sdr_file_pattern,
                     storage_options=dict(profile="someprofile"))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"))
    with patched_publisher() as messages:
        with patched_bucket_listener(records):
            minio_notification_watcher.file_publisher(fs_config=s3_config,
                                                      publisher_config=publisher_settings,
                                                      message_config=message_settings)
    message = Message(rawstr=messages[0])
    assert message.data["sensor"] == "viirs"
    assert message.data["platform_name"] == "npp"


def test_publish_paths_with_pattern_and_aliases(patched_bucket_listener):  # noqa
    """Test publishing paths."""
    s3_config = dict(endpoint_url="someendpoint",
                     bucket_name="viirs-data",
                     file_pattern=sdr_file_pattern,
                     storage_options=dict(profile="someprofile"))
    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/viirs/l1b/", atype="file", data=dict(sensor="viirs"),
                            aliases={"platform_name": {"npp": "Suomi-NPP"}})
    with patched_publisher() as messages:
       with patched_bucket_listener(records):
              minio_notification_watcher.file_publisher(fs_config=s3_config,
                                                        publisher_config=publisher_settings,
                                                        message_config=message_settings)
    message = Message(rawstr=messages[0])
    assert message.data["sensor"] == "viirs"
    assert message.data["platform_name"] == "Suomi-NPP"


records = \
[{"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:22.544Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470908962A15",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "9f6fbbf9e2b44509afa922b8d9884682",
                                 "key":
                                 "sdr/SVM13_npp_d20240408_t1006227_e1007469_b64498_c20240408102334392250_cspp_dev.h5",
                                 "sequencer": "17C447090F3AB797",
                                 "size": 22183568,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:22.666Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470912FB48E9",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "f92d4f4c1b33b8397fc710aed80ab134",
                                 "key":
                                 "sdr/SVM14_npp_d20240408_t1006227_e1007469_b64498_c20240408102334431798_cspp_dev.h5",
                                 "sequencer": "17C4470916824E6A",
                                 "size": 12357960,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:22.790Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C447091A7B932E",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "9ffffd85b277704d0b85ecc22d8337ce",
                                 "key":
                                 "sdr/SVM15_npp_d20240408_t1006227_e1007469_b64498_c20240408102334471520_cspp_dev.h5",
                                 "sequencer": "17C447091DE5AEA9",
                                 "size": 12357960,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:22.925Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470921D01846",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "a7cea3e72dea4845905922a851fac53c",
                                 "key":
                                 "sdr/SVM16_npp_d20240408_t1006227_e1007469_b64498_c20240408102334509150_cspp_dev.h5",
                                 "sequencer": "17C4470925EB97C5",
                                 "size": 12357960,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:CompleteMultipartUpload",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:25.377Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"content-length": "518",
                                    "x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C44709B7B42BA6",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "6accbb0d6937313fb06b5c09a88b0045-7",
                                 "key":
                                 "sdr/GIMGO_npp_d20240408_t1006227_e1007469_b64498_c20240408102307309287_cspp_dev.h5",
                                 "sequencer": "17C44709B802FBCF",
                                 "size": 324490832,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:CompleteMultipartUpload",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:27.829Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"content-length": "518",
                                    "x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470A4A03BE3D",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "36a554f5a2f344ed5f3ffb7826f0a17a-7",
                                 "key":
                                 "sdr/GITCO_npp_d20240408_t1006227_e1007469_b64498_c20240408102307123064_cspp_dev.h5",
                                 "sequencer": "17C4470A4A2E9DF0",
                                 "size": 324490848,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:28.227Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470A55A65AC6",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "c5804716bfa4431e35957308d0266b37",
                                 "key":
                                 "sdr/SVI01_npp_d20240408_t1006227_e1007469_b64498_c20240408102333369886_cspp_dev.h5",
                                 "sequencer": "17C4470A61DAC380",
                                 "size": 49222736,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:28.605Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470A6C179370",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "6e0465a1a67e401d08312d511e14b797",
                                 "key":
                                 "sdr/SVI02_npp_d20240408_t1006227_e1007469_b64498_c20240408102333521003_cspp_dev.h5",
                                 "sequencer": "17C4470A7877A613",
                                 "size": 49222736,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:28.998Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470A8318051B",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "f7a1a9089841e810a4249ee43ac7d344",
                                 "key":
                                 "sdr/SVI03_npp_d20240408_t1006227_e1007469_b64498_c20240408102333616593_cspp_dev.h5",
                                 "sequencer": "17C4470A8FCB6DF1",
                                 "size": 49222736,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:29.475Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470A99695857",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "ade56421a7d2b40a27ac32312d496c99",
                                 "key":
                                 "sdr/SVI04_npp_d20240408_t1006227_e1007469_b64498_c20240408102333718056_cspp_dev.h5",
                                 "sequencer": "17C4470AAC4C3A1E",
                                 "size": 49222736,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:29.881Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470AB6F25E72",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "70c7e8e51601a450c5cd73b3d6e2d57e",
                                 "key":
                                 "sdr/SVI05_npp_d20240408_t1006227_e1007469_b64498_c20240408102333815418_cspp_dev.h5",
                                 "sequencer": "17C4470AC47FA93E",
                                 "size": 49222736,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:CompleteMultipartUpload",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:31.138Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"content-length": "518",
                                    "x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470B0F539914",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "8cfd404cd0a9fde9088ebb8472c0eb7d-4",
                                 "key":
                                 "sdr/GDNBO_npp_d20240408_t1006227_e1007469_b64498_c20240408102306939475_cspp_dev.h5",
                                 "sequencer": "17C4470B0F75A541",
                                 "size": 168652400,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:Put",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:31.293Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470B143343A8",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "45a903f620151c6e84b3dcabc7662692",
                                 "key":
                                 "sdr/SVDNB_npp_d20240408_t1006227_e1007469_b64498_c20240408102332971066_cspp_dev.h5",
                                 "sequencer": "17C4470B18B25911",
                                 "size": 15662132,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]},
 {"Records": [{"awsRegion": "",
               "eventName": "s3:ObjectCreated:CompleteMultipartUpload",
               "eventSource": "minio:s3",
               "eventTime": "2024-04-08T10:24:32.244Z",
               "eventVersion": "2.0",
               "requestParameters": {"principalId": "someuser",
                                     "region": "",
                                     "sourceIPAddress": "172.29.4.164"},
               "responseElements": {"content-length": "518",
                                    "x-amz-id-2": "b675f5c02385af78c69266a96b22038bea3ecd7423814977a7267fd68ac2bbeb",
                                    "x-amz-request-id": "17C4470B5138A602",
                                    "x-minio-deployment-id": "a287d091-5b07-4b97-90c7-856aca7e9543",
                                    "x-minio-origin-endpoint": "https://172.31.28.90:443"},
               "s3": {"bucket": {"arn": "arn:aws:s3:::viirs-data",
                                 "name": "viirs-data",
                                 "ownerIdentity": {"principalId": "someuser"}},
                      "configurationId": "Config",
                      "object": {"contentType": "application/x-hdf5",
                                 "eTag": "fea57bce523e03c1cc38d864beb8a3d5-3",
                                 "key":
                                 "sdr/IVCDB_npp_d20240408_t1006227_e1007469_b64498_c20240408102333190566_cspp_dev.h5",
                                 "sequencer": "17C4470B515E6945",
                                 "size": 123520808,
                                 "userMetadata": {"content-type": "application/x-hdf5"}},
                      "s3SchemaVersion": "1.0"},
               "source": {"host": "172.29.4.164",
                          "port": "",
                          "userAgent": "aiobotocore/2.12.1 md/Botocore#1.34.51 "
                                       "ua/2.0 "
                                       "os/linux#4.18.0-513.18.1.el8_9.x86_64 "
                                       "md/arch#x86_64 lang/python#3.11.8 "
                                       "md/pyimpl#CPython "
                                       "cfg/retry-mode#legacy "
                                       "botocore/1.34.51"},
               "userIdentity": {"principalId": "someuser"}}]}]



"""Tests for the copernicus datapace watcher."""
import datetime

from freezegun import freeze_time
from posttroll.message import Message
from posttroll.testing import patched_publisher
from pytroll_watchers.dataspace_watcher import file_generator, file_publisher
from pytroll_watchers.testing import load_oauth_responses


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_dataspace_watcher():
    """Test the dataspace watcher."""
    check_interval = datetime.timedelta(0)  # run once and avoid the infinite loop
    filter_string = "contains(Name,'OL_1_EFR')"
    storage_options = dict(profile="mycopernicus")
    username = "user1"
    password = "pass1"  # noqa
    auth = dict(username=username, password=password)
    with load_oauth_responses("token", "filtered_yesterday"):
        files = list(file_generator(filter_string, check_interval, auth, datetime.timedelta(hours=24),
                                    storage_options))
        assert len(files) == 3
        s3path, metadata = files[0]
        assert s3path.storage_options == storage_options
        fname = "S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3"
        assert s3path.as_uri().endswith(fname)
        assert metadata["platform_name"] == "Sentinel-3B"
        assert metadata["sensor"] == "olci"
        assert metadata["start_time"] == datetime.datetime(2024, 4, 15, 7, 40, 29, 480000, tzinfo=datetime.timezone.utc)
        assert metadata["end_time"] == datetime.datetime(2024, 4, 15, 7, 43, 29, 480000, tzinfo=datetime.timezone.utc)
        assert metadata["boundary"]["coordinates"][0][0] == [67.8172, 69.3862]
        assert metadata["orbit_number"] == 31107


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_dataspace_watcher_without_storage_options():
    """Check that not providing storage options work."""
    check_interval = datetime.timedelta(0)
    filter_string = "contains(Name,'OL_1_EFR')"
    username = "user2"
    password = "pass2"  # noqa
    auth = dict(username=username, password=password)
    with load_oauth_responses("token", "filtered_yesterday"):
        files = list(file_generator(filter_string, check_interval, auth, start_from=datetime.timedelta(hours=24)))
        s3path, _ = files[0]
        assert s3path.storage_options == {}


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_dataspace_watcher_default_start_time():
    """Test that the dataspace watcher uses now as the default start time."""
    check_interval = datetime.timedelta(0)  # run once and avoid the infinite loop
    filter_string = "contains(Name,'OL_1_EFR')"
    username = "user3"
    password = "pass3"  # noqa
    auth = dict(username=username, password=password)
    with load_oauth_responses("token", "filtered_today"):
        files = list(file_generator(filter_string, check_interval, auth))
        assert len(files) == 0


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_dataspace_watcher_with_token_credentials_from_netrc(tmp_path):
    """Check that authentication can use netrc."""
    check_interval = datetime.timedelta(0)
    filter_string = "contains(Name,'OL_1_EFR')"
    netrc_host = "myitem"
    netrc_file = tmp_path / "netrc"

    with open(netrc_file, "w") as fd:
        fd.write(f"machine {netrc_host} login user@pytroll.org password mypassword")

    auth = dict(netrc_host=netrc_host, netrc_file=netrc_file)
    with load_oauth_responses("token", "filtered_yesterday"):
        files = list(file_generator(filter_string, check_interval, auth,
                                    start_from=datetime.timedelta(hours=24)))
        s3path, _ = files[0]
        assert s3path.storage_options == {}


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_dataspace_watcher_logs(caplog):
    """Check that logging works."""
    caplog.set_level("INFO")
    check_interval = datetime.timedelta(0)
    filter_string = "contains(Name,'OL_1_EFR')"
    username = "user4"
    password = "pass4"  # noqa
    auth = dict(username=username, password=password)
    with load_oauth_responses("token", "filtered_yesterday"):
        list(file_generator(filter_string, check_interval, auth, start_from=datetime.timedelta(hours=24)))
    assert "Finished polling." in caplog.text


@freeze_time(datetime.datetime.now(datetime.timezone.utc))
def test_publish_paths(caplog):
    """Test publishing paths."""
    filter_string = "contains(Name,'OL_1_EFR')"

    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l1b/", atype="file", data=dict(sensor="olci"))

    caplog.set_level("INFO")
    with patched_publisher() as messages:
        with load_oauth_responses("token", "filtered_yesterday"):
            ds_config = dict(filter_string=filter_string,
                            polling_interval=dict(minutes=0),
                            start_from=dict(hours=24),
                            dataspace_auth=dict(username="user5", password="pass5"),  # noqa
                            storage_options=dict(profile="someprofile"))


            file_publisher(fs_config=ds_config,
                           publisher_config=publisher_settings,
                           message_config=message_settings)
    assert "uri" not in message_settings["data"]
    assert len(messages) == 3
    message = Message(rawstr=messages[0])
    assert message.data["uri"] == "s3:///eodata/Sentinel-3/OLCI/OL_1_EFR___/2024/04/15/S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3"
    assert message.data["uid"] == "S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3"  # noqa
    assert message.data["sensor"] == "olci"
    assert message.data["filesystem"] == {"cls": "s3fs.core:S3FileSystem", "protocol": "s3", "args": [],
                                  "profile": "someprofile"}
    assert message.data["path"] == "/eodata/Sentinel-3/OLCI/OL_1_EFR___/2024/04/15/S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3"  # noqa
    assert f"Starting watch on dataspace for '{filter_string}'" in caplog.text



"""Common functions for publishing messages."""

import datetime
import json
import logging
from contextlib import closing, suppress
from copy import deepcopy

from posttroll.message import Message
from posttroll.publisher import create_publisher_from_dict_config
from trollsift import parse
from upath import UPath

logger = logging.getLogger(__name__)


class SecurityError(Exception):
    """An exception for breaking security rules."""


def file_publisher_from_generator(generator, publisher_config, message_config):
    """Publish files coming from local filesystem events.

    Args:
        generator: the generator to use for producing files. The generator must yield tuples of
            (filename, file_metadata).
        publisher_config: The configuration dictionary to pass to the posttroll publishing functions.
        message_config: The information needed to complete the posttroll message generation. Will be amended
            with the file metadata, and passed directly to posttroll's Message constructor.
            If it contains "archive_format", it is expected to have the archive type, and the contents of the archive
            will be published as a "dataset".

    Side effect:
        Publishes posttroll messages containing the location of the file with the following fields:
          - The "uid" which is the unique filename of the file.
          - The "uri" which provides an fsspec-style uri to the file. This does not however contain the connection
            parameters, or storage options of the filesystems referred to.
          - The "filesystem" which is the json-serialized filesystem information that can then be fed to fsspec.
          - The "path" which is the path of the file inside the filesystem.

        For example, the file
          - uid: `S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3/Oa02_radiance.nc`
          - uri: `s3:///eodata/Sentinel-3/OLCI/OL_1_EFR___/2024/04/15/S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3/Oa02_radiance.nc`
          - filesystem: `{"cls": "s3fs.core.S3FileSystem", "protocol": "s3", "args": [], "profile": "my_profile"}`
          - path: `/eodata/Sentinel-3/OLCI/OL_1_EFR___/2024/04/15/S3B_OL_1_EFR____20240415T074029_20240415T074329_20240415T094236_0179_092_035_1620_PS2_O_NR_003.SEN3/Oa02_radiance.nc`
    """  # noqa
    publisher = create_publisher_from_dict_config(publisher_config)
    publisher.start()
    archive_format = message_config.pop("archive_format", None)
    with closing(publisher):
        for file_item, file_metadata in generator:
            amended_message_config = deepcopy(message_config)
            amended_message_config.setdefault("data", {})

            if archive_format:
                dataset = [_build_file_location(unpacked_file) for unpacked_file in unpack(file_item, archive_format)]
                amended_message_config["data"]["dataset"] = dataset
            else:
                file_location = _build_file_location(file_item)
                amended_message_config["data"].update(file_location)

            aliases = amended_message_config.pop("aliases", {})
            apply_aliases(aliases, file_metadata)
            amended_message_config["data"].update(file_metadata)
            msg = Message(**amended_message_config)
            logger.info(f"Sending {str(msg)}")
            publisher.send(str(msg))


def unpack(path, archive_format):
    """Unpack the path and yield the extracted filenames."""
    import fsspec
    fs = fsspec.get_filesystem_class(archive_format)(fsspec.open(path.path, **path.storage_options))
    files = fs.find("/")
    for fi in files:
        yield UPath(fi,
                    protocol=archive_format,
                    target_protocol=path.protocol,
                    target_options=path.storage_options,
                    fo=path.as_uri())


def _build_file_location(file_item):
    file_location = dict()
    file_location["uri"] = file_item.as_uri()
    file_location["uid"] = file_item.name
    with suppress(AttributeError):
        file_location["filesystem"] = json.loads(file_item.fs.to_json())
        file_location["path"] = file_item.path
    return file_location


def apply_aliases(aliases, metadata):
    """Apply aliases to the metadata.

    Args:
        aliases: a dict containing dicts for each key to be aliases. For example
            `{"platform_name": {"npp": "Suomi-NPP}}` will replace the `platform_name` "npp" with "Suomi-NPP".
        metadata: the metadata to fix
    """
    for key, val in metadata.items():
        if key in aliases:
            metadata[key] = aliases[key].get(val, val)


def fix_times(info):
    """Fix times so that date and time components are combined, and start time is before end time."""
    if "start_date" in info:
        info["start_time"] = datetime.datetime.combine(info["start_date"].date(),
                                                       info["start_time"].time())
        if "end_date" not in info:
            info["end_date"] = info["start_date"]
        del info["start_date"]
    if "end_date" in info:
        info["end_time"] = datetime.datetime.combine(info["end_date"].date(),
                                                     info["end_time"].time())
        del info["end_date"]
    if "end_time" in info:
        while info["start_time"] > info["end_time"]:
            info["end_time"] += datetime.timedelta(days=1)

def parse_metadata(file_pattern, path):
    """Parse metadata from the filename."""
    if file_pattern is not None:
        file_metadata = parse(file_pattern, path)
        fix_times(file_metadata)
    else:
        file_metadata = {}
    return file_metadata



"""Backend for listening to local events. Uses Watchdog."""

import fnmatch
import os
from contextlib import contextmanager
from functools import partial
from queue import Queue

from trollsift import globify
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver


@contextmanager
def listen_to_local_events(directory, file_pattern=None, observer_type="os"):
    """Listen to local events.

    This context manager returns a generator producing filenames that are detected locally.

    Args:
        directory: The directory to watch for changes.
        file_pattern: the (trollsift) pattern to use globbing the events and filter them.
        observer_type: how to watch for events ("os" or "minio")

    Yields:
        A generator of filenames.
    """
    queue = Queue()
    def function_to_run(filename):
        queue.put(filename)

    extra_args = {}
    if file_pattern:
        extra_args["glob_pattern"] = globify(str(file_pattern))

    if observer_type == "os":
        obs = _create_watchdog_os_observer(str(directory), function_to_run, **extra_args)
    elif observer_type == "polling":
        obs = _create_watchdog_polling_observer(str(directory), function_to_run, **extra_args)
    else:
        raise ValueError("Observer type can be either 'os' or 'polling'.")

    obs.start()
    yield _iterate_over_queue(queue)
    obs.stop()


def _create_watchdog_polling_observer(directory, function_to_run, glob_pattern="*", timeout=1.0):
    """Create a watchdog polling observer on directory.

    Args:
        directory: the directory to watch for events.
        function_to_run: the function to run on detected files. Takes a single argument: the full path to the file.
        glob_pattern: the globbing pattern to use for filtering files. Defaults to "*"
        timeout: the timeout to use for polling, in seconds.

    Returns:
        The instanciated observer object.
    """
    observer_class = partial(PollingObserver, timeout=timeout)
    handler_class = _WatchdogCreationHandler
    return _create_watchdog_observer(directory, function_to_run, glob_pattern, observer_class, handler_class)


def _create_watchdog_os_observer(directory, function_to_run, glob_pattern="*", timeout=1.0):
    """Create a watchdog os-dependent observer on directory.

    Args:
        directory: the directory to watch for events.
        function_to_run: the function to run on detected files. Takes a single argument: the full path to the file.
        glob_pattern: the globbing pattern to use for filtering files. Defaults to "*"
        timeout: the timeout to use for detecting, in seconds.

    Returns:
        The instanciated observer object.
    """
    observer_class = partial(Observer, timeout=timeout, generate_full_events=True)
    handler_class = _WatchdogChangeHandler
    return _create_watchdog_observer(directory, function_to_run, glob_pattern, observer_class, handler_class)


def _iterate_over_queue(queue):
    """Iterate over the queue.

    This is it's own function so that it can be mocked during tests.
    """
    return iter(queue.get, None)


def _create_watchdog_observer(directory, function_to_run, glob_pattern, observer_class, handler_class):
    """Create a watchdog observer.

    Args:
        directory: the directory to watch for events.
        glob_pattern: the globbing pattern to use for filtering files
        function_to_run: the function to run on detected files. Takes a single argument: the full path to the file.
        observer_class: the class to instanciate as observer.
        handler_class: the class to use as handler.

    Returns:
        The instanciated observer object.
    """
    observer = observer_class()
    handler = handler_class(function_to_run, os.path.join(directory, glob_pattern))

    observer.schedule(handler, directory)

    return observer


class _WatchdogHandler(FileSystemEventHandler):
    """Trigger processing on filesystem events, with filename matching."""

    def __init__(self, fun, pattern=None):
        """Initialize the processor."""
        super().__init__()
        self.fun = fun
        self.pattern = pattern

    def dispatch(self, event):
        """Dispatches events to the appropriate methods."""
        if self.pattern is None:
            return super().dispatch(event)
        if event.is_directory:
            return
        if getattr(event, "dest_path", None):
            pathname = os.fsdecode(event.dest_path)
        elif event.src_path:
            pathname = os.fsdecode(event.src_path)
        if fnmatch.fnmatch(pathname, self.pattern):
            super().dispatch(event)


class _WatchdogChangeHandler(_WatchdogHandler):
    """Trigger processing on filesystem events that change a file (moving, close (write))."""

    def on_closed(self, event):
        """Process file closed."""
        self.fun(event.src_path)

    def on_moved(self, event):
        """Process a file being moved to the destination directory."""
        self.fun(event.dest_path)


class _WatchdogCreationHandler(_WatchdogHandler):
    """Trigger processing on filesystem events that create a file (moving, creation)."""

    def on_created(self, event):
        """Process file closing."""
        self.fun(event.src_path)

    def on_moved(self, event):
        """Process a file being moved to the destination directory."""
        self.fun(event.dest_path)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from pytroll_watchers.version import version

project = "pytroll-watchers"
copyright = "2024, Martin Raspaud"
author = "Martin Raspaud"
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.napoleon", "sphinx.ext.autodoc", "sphinx.ext.intersphinx"]
autodoc_mock_imports = ["watchdog", "minio", "posttroll", "pytest", "trollsift", "universal_path",
                        "freezegun", "responses", "oauthlib", "requests_oauthlib", "defusedxml", "redis"]

templates_path = ["_templates"]
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]

# intersphinx
intersphinx_mapping = {
    "posttroll": ("https://posttroll.readthedocs.io/en/latest/", None),
    "redis": ("https://redis.readthedocs.io/en/latest/", None),
}



"""Watcher for non-remote file systems.

Either using OS-based envents (like inotify on linux), or polling.
"""
import logging
import os
from pathlib import Path
from urllib.parse import urlunparse

from upath import UPath
from upath._flavour import WrappedFileSystemFlavour

from pytroll_watchers.backends.local import listen_to_local_events
from pytroll_watchers.publisher import SecurityError, file_publisher_from_generator, parse_metadata

# This is a workaround for a but in universal_pathlib, see
WrappedFileSystemFlavour.protocol_config["netloc_is_anchor"].add("ssh")
logger = logging.getLogger(__name__)


def file_publisher(fs_config, publisher_config, message_config):
    """Publish files coming from local filesystem events.

    Args:
        fs_config: the configuration for the filesystem watching, will be passed as argument to `file_generator`.
        publisher_config: The configuration dictionary to pass to the posttroll publishing functions.
        message_config: The information needed to complete the posttroll message generation. Will be amended
             with the file metadata, and passed directly to posttroll's Message constructor.
    """
    logger.info(f"Starting watch on '{fs_config['directory']}'")
    if "password" in fs_config.get("storage_options", []):
        raise SecurityError("A password cannot be published safely.")
    generator = file_generator(**fs_config)
    return file_publisher_from_generator(generator, publisher_config, message_config)


def file_generator(directory, observer_type="os", file_pattern=None, protocol=None, storage_options=None):
    """Generate new files appearing in the watched directory.

    Args:
        directory: The locally accessible directory to watch for changes.
        observer_type: What to use for detecting changes.
            It can be either "os" for os-based detections (eg inotify on linux, but can be polling on some os's),
            "polling" for detecting changes through polling, or the actual watchdog class to use as observer.
            Defaults to "os".
        file_pattern: The trollsift pattern to use for matching and extracting metadata from the filename.
            This must not include the directory.
        protocol (optional): In case the file has to be advertised with another protocol than "file".
        storage_options: The storage options for the other protocol. Will be ignored if protocol is None.

    Returns:
        A tuple of Path or UPath and file metadata.

    Examples:
        To iterate over new files in `/tmp/`:

        >>> for filename in file_generator("/tmp/", file_pattern="{start_time:%Y%m%d_%H%M}_{product}.tif")
        ...    print(filename)
        Path("/tmp/20200428_1000_foo.tif")


        To get UPaths with ssh as protocol and a specific host:

        >>> for filename in file_generator("/tmp/", file_pattern="{start_time:%Y%m%d_%H%M}_{product}.tif",
        ...                                protocol="ssh", storage_option=dict(host="myhost.pytroll.org"))
        UPath("ssh:///tmp/20200428_1000_foo.tif")  # .storage_options will show the host.

    """
    file_metadata = {}
    pattern = os.path.join(directory, file_pattern) if file_pattern is not None else None
    with listen_to_local_events(directory, file_pattern, observer_type) as events:
        for path in events:
            try:
                file_metadata = parse_metadata(pattern, path)
            except ValueError:
                continue
            if protocol is not None:
                uri = urlunparse((protocol, None, str(path), None, None, None))
                yield UPath(uri, **storage_options), file_metadata
            else:
                yield Path(path), file_metadata



"""Tests for the publisher functionality."""

import os
from shutil import make_archive

from posttroll.testing import patched_publisher
from pytroll_watchers.publisher import file_publisher_from_generator
from upath import UPath


def test_unpacking(tmp_path):
    """Test unpacking the watched file's components to a dataset message."""
    file1 = tmp_path / "to_zip" / "file1"
    file2 = tmp_path / "to_zip" / "file2"
    zip_file = tmp_path / "archived.zip"

    file1.parent.mkdir()

    open(file1, "a").close()
    open(file2, "a").close()

    make_archive(os.path.splitext(zip_file)[0], "zip", tmp_path / "to_zip")

    path = UPath("file://" + str(zip_file))

    publisher_settings = dict(nameservers=False, port=1979)
    message_settings = dict(subject="/segment/olci/l2/", atype="dataset", data=dict(sensor="olci"),
                            archive_format="zip")

    with patched_publisher() as messages:
        file_publisher_from_generator([[path, dict()]],
                                      publisher_config=publisher_settings,
                                      message_config=message_settings)

    assert "zip://file1" in messages[0]
    assert "zip://file2" in messages[0]
